class NodeMetaclass(type):
    def __new__(cls, name, bases, attrs):
        if bases:
            base = bases[0]
            attributes = base.attributes.copy()
        else:
            attributes = {}
        attributes.update(
            {
                key: value
                for key, value in attrs.items()
                if not key.startswith("_") and key != "attributes"
            }
        )
        attrs["attributes"] = attributes
        return super().__new__(cls, name, bases, attrs)


class Node(metaclass=NodeMetaclass):
    attributes = []

    def __init__(self, **kwargs):
        for attr_name, value in kwargs.items():
            if attr_name not in self.attributes:
                raise RuntimeError("Invalid attribute {}".format(attr_name))
            setattr(self, attr_name, value)


class Const(Node):
    value = None


class StrConst(Const):
    pass


class Var(Node):
    name = None


class Value(Node):
    body = None


class ToStr(Node):
    value = None


class BaseObjNode(Node):
    obj = None


class GetAttr(BaseObjNode):
    attr_name = None


class GetItem(BaseObjNode):
    key = None


class Arg(Node):
    value = None


class Kwarg(Node):
    name = None
    value = None


class Call(BaseObjNode):
    args = None


class Filter(BaseObjNode):
    name = None
    args = None


class Template(Node):
    body = None


class CellOutput(Node):
    base_cell = None
    value = None
    style = None
    row_height = None
    col_width = None
    merge = None

    def adjust(self, row, col):
        self.base_cell = (self.base_cell[0] + row, self.base_cell[1] + col)


class CellGroup(Node):
    base_cell = None
    last_cell = None
    body = None

    def adjust(self, row, col):
        self.base_cell = (self.base_cell[0] + row, self.base_cell[1] + col)
        self.last_cell = (self.last_cell[0] + row, self.last_cell[1] + col)

    @property
    def height(self):
        return self.last_cell[0] - self.base_cell[0] + 1

    @property
    def width(self):
        return self.last_cell[1] - self.base_cell[1] + 1

    def get_cell_range(self):
        for row in range(self.base_cell[0], self.last_cell[0] + 1):
            for col in range(self.base_cell[1], self.last_cell[1] + 1):
                yield (row, col)


class Remove(CellGroup):
    pass


class If(CellGroup):
    condition = None
    else_block = None


class Sheet(CellGroup):
    name = None

    def __init__(self, name, max_row, max_col, body):
        super().__init__(
            base_cell=(0, 0), last_cell=(max_row - 1, max_col - 1), body=body
        )
        self.name = name


class CellLoop(CellGroup):
    target = None
    items = None
    name = None
    body = None
    direction = None


class SheetLoop(CellGroup):
    target = None
    items = None
    name = None
    sheet = None


class FuncArg(Node):
    start_index = None
    end_index = None
    cells = None
    direction = None


class Merge(Node):
    rows = None
    cols = None


class FuncArgDirection(Node):
    direction = None


class FuncCellOutput(CellOutput):
    default_value = None
    args = None

    def adjust(self, row, col):
        for arg in self.args:
            arg.cells = [
                (cell[0] - self.base_cell[0], cell[1] - self.base_cell[1])
                for cell in arg.cells
            ]
        super().adjust(row, col)


class ColWidth(Node):
    value = None


class RowHeight(Node):
    value = None



import string

import pyparsing as pp
from pyparsing import pyparsing_common as ppc

from . import nodes, utils


LPAR, RPAR, LBRACK, RBRACK, DOT, EQ, COMMA, COLON = [pp.Suppress(_) for _ in "()[].=,:"]


hex_constant = pp.Regex(r"0[xX][0-9a-fA-F]+").addParseAction(
    lambda t: int(t[0][2:], 16)
)
int_constant = hex_constant | ppc.integer
double_constant = ppc.real


def double_constant_pa(r):
    return nodes.Const(value=float(r.asList()[0]))


double_constant.setParseAction(double_constant_pa)
string_constant = pp.quotedString


def string_constant_pa(r):
    res = r.asList()[0]
    return nodes.StrConst(value=res)


string_constant.setParseAction(string_constant_pa)

boolean_constant = (
    pp.Keyword("true") | pp.Keyword("True") | pp.Keyword("false") | pp.Keyword("False")
)


def boolean_constant_pa(r):
    res = r.asList()[0].lower()
    if res == "true":
        return nodes.Const(value=True)
    else:
        return nodes.Const(value=False)


boolean_constant.setParseAction(boolean_constant_pa)


constant = double_constant | int_constant | string_constant | boolean_constant
constant.setResultsName("constant")


def constant_pa(r):
    return nodes.Const(value=r.asList()[0])


int_constant.setParseAction(constant_pa)


name = pp.Word(pp.alphas + "_", pp.alphanums + "_")
get_attr = DOT + name("attr_name")


def get_attr_pa(r):
    return nodes.GetAttr(attr_name=r.attr_name)


get_attr.setParseAction(get_attr_pa)

expr = pp.Forward()
get_item = LBRACK + expr("key") + RBRACK


def get_item_pa(r):
    return nodes.GetItem(key=r.key)


get_item.setParseAction(get_item_pa)

kwarg = name("name") + EQ + expr("value")


def kwarg_pa(r):
    return nodes.Kwarg(name=r.name, value=r.value)


kwarg.setParseAction(kwarg_pa)

arg = expr("arg")


def arg_pa(r):
    return nodes.Arg(value=r.asList()[0])


arg.setParseAction(arg_pa)

call_args = pp.delimitedList(kwarg | arg)
call = LPAR + pp.Group(pp.Optional(call_args))("args") + RPAR


def call_pa(r):
    return nodes.Call(args=r.asList()[0])


call.setParseAction(call_pa)

var = name("var")


def var_pa(r):
    return nodes.Var(name=r.asList()[0])


var.setParseAction(var_pa)

expr << ((constant | var) + pp.ZeroOrMore(get_attr | get_item | call))


def expr_pa(r):
    obj = None
    r = r.asList()
    for part in r:
        if obj is not None:
            part.obj = obj
        obj = part
    return obj


expr.setParseAction(expr_pa)

filter_expr = name("name") + pp.Optional(
    LPAR + pp.Group(pp.Optional(call_args))("args") + RPAR
)


def filter_expr_pa(r):
    return nodes.Filter(name=r.name, args=r.args)


filter_expr.setParseAction(filter_expr_pa)

expr_with_filter = expr + pp.ZeroOrMore(pp.Suppress("|") + filter_expr)


expr_with_filter.setParseAction(expr_pa)


def parse_expr_with_filter(s):
    return expr_with_filter.parseString(s, True).asList()[0]


cell = pp.Word(pp.alphas.upper())("col") + pp.Word(pp.nums)("row")


def cell_pa(r):
    res = (int(r["row"]), utils.col_str_to_int(r["col"]))
    return res


cell.setParseAction(cell_pa)


cell_range = cell("start_cell") + COLON + cell("end_cell")


def cell_range_pa(r):
    return r.asList()


cell_range.setParseAction(cell_range_pa)


def param_pa(r):
    r = r.asList()
    res = {"name": r[0], "value": r[1:]}
    if len(res["value"]) == 1:
        res["value"] = res["value"][0]
    return res


def make_param(param_name, param_value):
    r = pp.Keyword(param_name) + EQ + param_value
    r.setParseAction(param_pa)
    return r


for_statement = (
    pp.Keyword("for").suppress()
    + name("target")
    + pp.Keyword("in").suppress()
    + expr("items")
)


def for_statement_pa(r):
    return {"target": r["target"], "items": r["items"]}


for_statement.setParseAction(for_statement_pa)

cell_loop = for_statement("stmt") + pp.Group(
    pp.Optional(
        COMMA
        + pp.delimitedList(make_param("name", name) | make_param("last_cell", cell))
    )
)("params")


def cell_loop_pa(r):
    keywords = {param["name"]: param["value"] for param in r["params"].asList()}
    return nodes.CellLoop(
        target=r["stmt"]["target"], items=r["stmt"]["items"], **keywords
    )


cell_loop.setParseAction(cell_loop_pa)


def parse_cell_loop_directive(s):
    return cell_loop.parseString(s, True).asList()[0]


sheet_loop = for_statement("stmt") + pp.Group(
    pp.Optional(COMMA + pp.delimitedList(make_param("name", name)))
)("params")


def sheet_loop_pa(r):
    keywords = {param["name"]: param["value"] for param in r["params"].asList()}
    return nodes.SheetLoop(
        target=r["stmt"]["target"], items=r["stmt"]["items"], **keywords
    )


sheet_loop.setParseAction(sheet_loop_pa)


def parse_sheet_loop_directive(s):
    return sheet_loop.parseString(s, True).asList()[0]


merge = pp.delimitedList(make_param("rows", expr) | make_param("cols", expr))


def merge_pa(r):
    return nodes.Merge(**{param["name"]: param["value"] for param in r.asList()})


merge.setParseAction(merge_pa)


def parse_merge(s):
    return merge.parseString(s, True).asList()[0]


group = pp.delimitedList(make_param("last_cell", cell))


def group_pa(r):
    return nodes.CellGroup(**{param["name"]: param["value"] for param in r.asList()})


group.setParseAction(group_pa)


def parse_group(s):
    return group.parseString(s, True).asList()[0]


remove = pp.Optional(make_param("last_cell", cell))


def remove_pa(r):
    params = {param["name"]: param["value"] for param in r.asList()}
    return nodes.Remove(**params)


remove.setParseAction(remove_pa)


def parse_remove(s):
    return remove.parseString(s, True).asList()[0]


excel_func_arg = cell_range | cell


def excel_func_arg_pa(r):
    return r.asList()


excel_func_arg.setParseAction(excel_func_arg_pa)


excel_func_args = pp.ZeroOrMore(excel_func_arg)


def parse_func_args(s):
    return [
        ((start_index, end_index), parse_res)
        for (parse_res, start_index, end_index) in excel_func_args.scanString(s)
    ]


if_d = pp.delimitedList(
    make_param("condition", expr)
    | make_param("last_cell", cell)
    | make_param("else", excel_func_arg)
)


def if_pa(r):
    params = {param["name"]: param["value"] for param in r.asList()}
    params["else_block"] = params.pop("else", None)
    return nodes.If(**params)


if_d.setParseAction(if_pa)


def parse_if(s):
    return if_d.parseString(s, True).asList()[0]


def parse_col_width(s):
    return nodes.ColWidth(value=parse_expr_with_filter(s))


def parse_row_height(s):
    return nodes.RowHeight(value=parse_expr_with_filter(s))



import copy
import io

import pyparsing
from openpyxl import load_workbook, styles

from . import grammar, nodes, utils
from .consts import LoopDirection, FuncArgDirection
from .exceptions import ParseError


class Parser:
    def __init__(self, source, filename=None):
        self.wb = load_workbook(io.BytesIO(source))
        self.source_hint = ["filename:{}".format(filename)] if filename else []

    def parse(self):
        template = nodes.Template(body=[])
        self.styles = {}
        for sheet_name in self.wb.sheetnames:
            self.source_hint.append("sheet:{}".format(sheet_name))
            ws = self.wb[sheet_name]
            self.cells = {}
            self.merged_cells = {}
            self.post_remove = []
            self.original_cell_groups = {}
            for cr in ws.merged_cells:
                self.merged_cells[(cr.min_row, cr.min_col)] = (
                    cr.max_row - cr.min_row + 1,
                    cr.max_col - cr.min_col + 1,
                )
            for row in range(1, ws.max_row + 1):
                for col in range(1, ws.max_column + 1):
                    self.source_hint.append(
                        "cell:{}".format(utils.cell_int_to_str(row, col))
                    )
                    node = self.parse_cell(ws, row, col)
                    self.source_hint.pop()
                    self.cells[(row, col)] = node
            directives = self.get_directives(ws)
            self.directives = self.parse_directives(directives)
            body = self._process_cell_group(1, 1, ws.max_row, ws.max_column)
            self.source_hint.append("sheet name")
            sheet_name_node = self.parse_value(sheet_name)
            self.source_hint.pop()
            if isinstance(body[0], nodes.SheetLoop):
                root_node = body[0]
                root_node.sheet.name=nodes.ToStr(value=sheet_name_node)
            else:
                root_node = nodes.Sheet(
                    name=nodes.ToStr(value=sheet_name_node),
                    max_row=ws.max_row,
                    max_col=ws.max_column,
                    body=body,
                )
            self.source_hint.pop()

            assert not self.source_hint
            if self.post_remove:
                for start_cell, end_cell in self.post_remove:
                    target_cell_group = None
                    target_cell = None
                    for (row, col), cell_group in self.original_cell_groups.items():
                        if (
                            start_cell[0] >= row
                            and start_cell[1] >= col
                            and end_cell[0] <= row + cell_group.height
                            and end_cell[1] <= col + cell_group.width
                        ):
                            target_cell_group = cell_group
                            target_cell = (row, col)
                    assert target_cell_group is not None
                    target_cell_group.body.append(
                        nodes.Remove(
                            base_cell=(
                                start_cell[0] - target_cell[0],
                                start_cell[1] - target_cell[1],
                            ),
                            last_cell=(
                                end_cell[0] - target_cell[0],
                                end_cell[1] - target_cell[1],
                            ),
                        )
                    )
            template.body.append(root_node)
            del self.post_remove
            del self.original_cell_groups
        return template, list(self.styles.values())

    def _process_cell_group(self, start_row, start_col, end_row, end_col):
        body = []
        for row in range(start_row, end_row + 1):
            for col in range(start_col, end_col + 1):
                if (row, col) in self.directives and self.directives[(row, col)]:
                    cur_directives = self.directives[(row, col)]
                    cur_directive = cur_directives.pop(0)
                    if isinstance(cur_directive, nodes.SheetLoop):
                        cur_directive.last_cell = (end_row, end_col)
                    if not cur_directives:
                        del self.directives[(row, col)]
                    method = getattr(
                        self,
                        "_process_{}".format(cur_directive.__class__.__name__.lower()),
                    )
                    cur_directive = method(cur_directive)
                    if cur_directive is not None:
                        if isinstance(cur_directive, nodes.CellGroup):
                            self.original_cell_groups[(row, col)] = cur_directive
                        cur_directive.adjust(-start_row, -start_col)
                        body.append(cur_directive)
                if (row, col) in self.cells:
                    cell = self.cells.pop((row, col))
                    cell.adjust(-start_row, -start_col)
                    body.append(cell)
        return body

    def _process_cellloop(self, cell_loop):
        cell_loop.body = self._process_cell_group(
            cell_loop.base_cell[0],
            cell_loop.base_cell[1],
            cell_loop.last_cell[0],
            cell_loop.last_cell[1],
        )
        return cell_loop

    def parse_directives(self, directives):
        res_directives = {}
        for (row, col), directives in directives.items():
            self.source_hint.append("cell:{}".format(utils.cell_int_to_str(row, col)))
            res_directives[(row, col)] = []
            for index, directive in enumerate(directives):
                self.source_hint.append("directive_index:{}".format(index))
                res_directives[(row, col)].append(
                    self.parse_directive(row, col, directive)
                )
                self.source_hint.pop()
            self.source_hint.pop()
        return res_directives

    def parse_directive(self, row, col, directive):
        if "," in directive:
            directive_name, tail = directive.split(",", 1)
        else:
            directive_name, tail = directive, ""
        directive_name = directive_name.lower().replace("-", "_")
        method = getattr(self, "parse_{}".format(directive_name), None)
        if method is None:
            raise ParseError(
                "Invalid directive {}".format(directive_name), self.source_hint
            )
        node = method(tail)
        node.base_cell = (row, col)
        if hasattr(node, "last_cell"):
            if node.last_cell is None:
                node.last_cell = node.base_cell
            if len(node.last_cell) == 1:
                node.last_cell = node.last_cell[0]
        return node

    def parse_merge(self, directive_def):
        return self._parse_pp(grammar.parse_merge, directive_def)

    def _process_merge(self, merge):
        cell_output = self.cells[merge.base_cell]
        cell_output.merge = merge

    def parse_group(self, directive_def):
        return self._parse_pp(grammar.parse_group, directive_def)

    def parse_remove(self, directive_def):
        return self._parse_pp(grammar.parse_remove, directive_def)

    def _process_remove(self, remove):
        for cell in remove.get_cell_range():
            self.directives.pop(cell, None)
            self.cells.pop(cell, None)
        return remove

    def parse_if(self, directive_def):
        return self._parse_pp(grammar.parse_if, directive_def)

    def parse_col_width(self, directive_def):
        return self._parse_pp(grammar.parse_col_width, directive_def)

    def parse_row_height(self, directive_def):
        return self._parse_pp(grammar.parse_row_height, directive_def)

    def _process_colwidth(self, col_width):
        self.cells[col_width.base_cell].col_width = col_width.value

    def _process_rowheight(self, row_height):
        self.cells[row_height.base_cell].row_height = row_height.value

    def _process_if(self, if_d):
        if_d.body = self._process_cell_group(
            if_d.base_cell[0], if_d.base_cell[1], if_d.last_cell[0], if_d.last_cell[1]
        )
        if if_d.else_block:
            start_cell = if_d.else_block[0]
            if len(if_d.else_block) == 2:
                end_cell = if_d.else_block[1]
            else:
                end_cell = start_cell
            if_d.else_block = self._process_cell_group(
                start_cell[0], start_cell[1], end_cell[0], end_cell[1]
            )
            self.post_remove.append((start_cell, end_cell))
        return if_d

    def _process_cellgroup(self, cell_group):
        cell_group.body = self._process_cell_group(
            cell_group.base_cell[0],
            cell_group.base_cell[1],
            cell_group.last_cell[0],
            cell_group.last_cell[1],
        )
        return cell_group

    def _process_sheetloop(self, sheet_loop):
        sheet = nodes.Sheet(
            None,
            max_row=sheet_loop.last_cell[0],
            max_col=sheet_loop.last_cell[1],
            body=self._process_cell_group(
                sheet_loop.base_cell[0],
                sheet_loop.base_cell[1],
                sheet_loop.last_cell[0],
                sheet_loop.last_cell[1],
            ),
        )
        sheet_loop.sheet = sheet
        return sheet_loop

    def parse_func_arg_v(self, directive_def):
        return nodes.FuncArgDirection(direction=FuncArgDirection.VERTICAL)

    def parse_func_arg_h(self, directive_def):
        return nodes.FuncArgDirection(direction=FuncArgDirection.HORIZONTAL)

    def _process_funcargdirection(self, func_arg_dir):
        cell = (func_arg_dir.base_cell[0], func_arg_dir.base_cell[1])
        func_cell = self.cells.get(cell)
        assert isinstance(func_cell, nodes.FuncCellOutput), "Cell must contain function"
        for arg in func_cell.args:
            arg.direction = func_arg_dir.direction
        assert (
            self.directives.get(cell) is None
        ), "Func-arg-direction directive must be the last"

    def parse_loop_right(self, directive_def):
        return self._parse_cell_loop(directive_def, LoopDirection.RIGHT)

    def parse_loop_down(self, directive_def):
        return self._parse_cell_loop(directive_def, LoopDirection.DOWN)

    def _parse_cell_loop(self, directive_def, loop_direction):
        node = self._parse_pp(grammar.parse_cell_loop_directive, directive_def)
        node.direction = loop_direction
        if node.last_cell is None:
            node.last_cell = node.base_cell
        return node

    def parse_loop_sheet(self, directive_def):
        node = self._parse_pp(grammar.parse_sheet_loop_directive, directive_def)
        return node

    def get_directives(self, ws):
        directives = {}
        for row in range(1, ws.max_row + 1):
            for col in range(1, ws.max_column + 1):
                cell = ws.cell(row, col)
                if cell.comment and cell.comment.text:
                    lines = [line.strip() for line in cell.comment.text.splitlines()]
                    lines = [line for line in lines if line]
                    if lines[0].lower() == "synt-v2":
                        v2_directives = self.get_directives_synt_v2(lines[1:])
                        for (row, col), cell_directives in v2_directives.items():
                            if (row, col) not in directives:
                                directives[(row, col)] = []
                            directives[(row, col)].extend(cell_directives)
                    else:
                        directives[(row, col)] = lines
        return directives

    def get_directives_synt_v2(self, lines):
        state = 0
        current_cell = None
        res_directives = {}
        for line in lines:
            if state == 0:
                current_cell = grammar.cell.parseString(line, True)[0]
                state = 1
            else:
                if all(c == "=" for c in line):
                    state = 0
                else:
                    if current_cell not in res_directives:
                        res_directives[current_cell] = []
                    res_directives[current_cell].append(line)
        return res_directives

    def _parse_pp(self, pp, s):
        try:
            return pp(s)
        except pyparsing.ParseBaseException as e:
            msg = "Error parse '{}'".format(s)
            raise ParseError(msg, self.source_hint, e)

    def parse_expression(self, expr):
        node = self._parse_pp(grammar.parse_expr_with_filter, expr)
        return node

    def parse_value(self, value):
        if not isinstance(value, str) or "{{" not in value:
            return nodes.Const(value=value)
        else:
            body = []
            index = 0
            while index < len(value):
                start_var_index = value.find("{{", index)
                # If value start with {{
                if start_var_index - index == 0:
                    end_var_index = value.find("}}", start_var_index)
                    if end_var_index != -1:
                        self.source_hint.append(
                            "index:{}-{}".format(start_var_index + 2, end_var_index)
                        )
                        body.append(
                            self.parse_expression(
                                value[start_var_index + 2 : end_var_index]
                            )
                        )
                        self.source_hint.pop()
                        index = end_var_index + 2
                    else:
                        body.append(nodes.Const(value=value[index:]))
                        index = len(value)
                else:
                    body.append(nodes.Const(value=value[:start_var_index]))
                    index = start_var_index
            if len(body) != 1:
                body = [
                    nodes.ToStr(value=child)
                    if not isinstance(child, nodes.Const)
                    else child
                    for child in body
                ]
            return nodes.Value(body=body)

    def parse_cell(self, ws, row, col):
        cell = ws.cell(row, col)
        style_name = None
        if cell.has_style:
            style_key = (
                copy.copy(cell.font),
                copy.copy(cell.fill),
                copy.copy(cell.border),
                copy.copy(cell.alignment),
                copy.copy(cell.number_format),
                copy.copy(cell.protection),
            )
            if style_key not in self.styles:
                new_style_name = "style{}".format(len(self.styles))
                new_style = styles.NamedStyle(new_style_name)
                new_style.font = style_key[0]
                new_style.fill = style_key[1]
                new_style.border = style_key[2]
                new_style.alignment = style_key[3]
                new_style.number_format = style_key[4]
                new_style.protection = style_key[5]
                self.styles[style_key] = new_style
                style_name = new_style_name
            else:
                style_name = self.styles[style_key].name
        col_letter = utils.col_int_to_str(col)
        col_width = ws.column_dimensions[col_letter].width
        col_index = col
        while not col_width:
            col_letter = utils.col_int_to_str(col_index)
            col_width = ws.column_dimensions[col_letter].width
            col_index -= 1
        node_kwargs = {
            "base_cell": (row, col),
            "style": style_name,
            "row_height": nodes.Const(value=ws.row_dimensions[row].height),
            "col_width": nodes.Const(value=col_width),
        }
        node_class = nodes.CellOutput
        if cell.value:
            if isinstance(cell.value, str) and cell.value.startswith("="):
                value = cell.value
                node_kwargs["args"] = self.parse_func_args(value)
                node_class = nodes.FuncCellOutput
            else:
                value = self.parse_value(cell.value)
        else:
            value = None
        node_kwargs["value"] = value
        if (row, col) in self.merged_cells:
            node_kwargs["merge"] = nodes.Merge(
                rows=nodes.Const(value=self.merged_cells[(row, col)][0]),
                cols=nodes.Const(value=self.merged_cells[(row, col)][1]),
            )
        cell_output = node_class(**node_kwargs)
        return cell_output

    def parse_func_args(self, s):
        res = []
        for (start_index, end_index), cell_def in grammar.parse_func_args(s):
            if len(cell_def) == 1:
                cells = cell_def
            else:
                cells = [
                    (row, col)
                    for row in range(cell_def[0][0], cell_def[1][0] + 1)
                    for col in range(cell_def[0][1], cell_def[1][1] + 1)
                ]
            res.append(
                nodes.FuncArg(start_index=start_index, end_index=end_index, cells=cells)
            )
        return res



import tempfile
import os

from xlsx_template.parser import Parser
from xlsx_template.code_generator import CodeGenerator
from xlsx_template.environment import Environment
from xlsx_template.runtime.context import Context


class Template:
    def __init__(self, source, env=None, debug=False):
        if env is None:
            env = Environment()
        self.env = env
        if hasattr(source, "read"):
            source = source.read()
        parser = Parser(source)
        self.root_node, self.styles = parser.parse()
        code_source = CodeGenerator().generate(self.root_node)
        if debug:
            fd, self.debug_file_name = tempfile.mkstemp(suffix=".py")
            filename = self.debug_file_name
            with open(fd, "w") as f:
                f.write(code_source)
            self.code_source = code_source
        else:
            filename = "<template>"
        source = compile(code_source, filename, "exec")
        self.namespace = {}
        exec(source, self.namespace)

    def render(self, context_data):
        return self.namespace["root"](
            Context(context_data, self.env), self.styles, self.env
        )



class Context:
    def __init__(self, context_data, env):
        self.context_data = context_data
        self.env = env

    def resolve(self, name):
        if name in self.context_data:
            return self.env.resolve(self.context_data[name], name, True)
        else:
            return self.env.resolve(None, name, False)



from .exceptions import TemplateRuntimeException
from . import filters
from .parser import Parser


class StrictResolveStrategy:
    def resolve(self, obj, name, found):
        if not found:
            raise Unresolved("Can not find variable '{}'".format(name))
        return obj


class StrictGetAttrStrategy:
    def get_attr(self, obj, attr_name):
        try:
            return getattr(obj, attr_name)
        except AttributeError as e:
            raise TemplateRuntimeException(
                "Can not get attrib {} from obj {}".format(attr_name, obj), e
            )


class SoftGetAttrStrategy:
    def get_attr(self, obj, attr_name):
        exc = None
        try:
            return getattr(obj, attr_name)
        except AttributeError as e:
            exc = e
        try:
            return obj[attr_name]
        except Exception as e:
            exc = exc or e
            raise TemplateRuntimeException(
                "Can not get attrib {} from obj {}".format(attr_name, obj), exc
            )


class StrictGetItemStrategy:
    def get_item(self, obj, key):
        try:
            return obj[key]
        except Exception as e:
            raise TemplateRuntimeException(
                "Can not get item {} from obj {}".format(key, obj), e
            )


class Environment:

    filters = {"default_if_none": filters.default_if_none, "yes_no": filters.yes_no}

    def __init__(
        self, resolve_strategy=None, get_attr_strategy=None, get_item_strategy=None
    ):
        if resolve_strategy is None:
            resolve_strategy = StrictResolveStrategy()

        if get_attr_strategy is None:
            get_attr_strategy = SoftGetAttrStrategy()

        if get_item_strategy is None:
            get_item_strategy = StrictGetItemStrategy()

        self.resolve_strategy = resolve_strategy
        self.get_attr_strategy = get_attr_strategy
        self.get_item_strategy = get_item_strategy

    def resolve(self, obj, name, found):
        return self.resolve_strategy.resolve(obj, name, found)

    def get_attr(self, obj, attr_name):
        return self.get_attr_strategy.get_attr(obj, attr_name)

    def get_item(self, obj, key):
        return self.get_item_strategy.get_item(obj, key)



class LoopContext:
    def __init__(self, items):
        if not hasattr(items, "__len__"):
            items = list(items)
        self.items = items
        self.index0 = -1
        self.length = len(items)

    def __iter__(self):
        for item in self.items:
            self.index0 += 1
            yield item

    @property
    def index(self):
        return self.index0 + 1

    @property
    def revindex(self):
        return self.length - self.index0

    @property
    def revindex0(self):
        return self.length - self.index

    @property
    def first(self):
        return self.index0 == 0

    @property
    def last(self):
        return self.index0 == self.length - 1



from collections import defaultdict, namedtuple
import itertools
import typing

from cached_property import cached_property

from .. import utils, consts


Size = namedtuple("Size", "height,width")


class Cell:
    def __init__(self, row, col, style, value, row_height, col_width):
        self.row = row
        self.col = col
        self.value = value
        self.style = style
        self.row_height = row_height
        self.col_width = col_width

    def move(self, row, col):
        self.row += row
        self.col += col

    def __eq__(self, other):
        return (
            self.row == other.row
            and self.col == other.col
            and self.value == other.value
            and self.style == other.style
        )

    def __str__(self):  # pragma: no cover
        return "Cell({}, {}, {})".format(self.row, self.col, self.value)

    def __repr__(self):  # pragma: no cover
        return str(self)


class Merge:
    def __init__(self, row, col, rows, cols):
        self.row, self.col = row, col
        self.rows, self.cols = rows, cols

    def __str__(self):
        return "{}{}:{}{}"

    def move(self, row, col):
        self.row += row
        self.col += col


class FuncArg:
    def __init__(self, start_index, end_index, cells, direction=None):
        self.start_index = start_index
        self.end_index = end_index
        self.cells = cells
        self.final_cells = []
        self.direction = direction

    def finalize_cells(self, current_row, current_col, initial_cell, final_cells):
        self.cells.remove(initial_cell)
        if self.direction is not None:
            if self.direction == consts.FuncArgDirection.HORIZONTAL:
                cells = [
                    (cell.row, cell.col)
                    for cell in final_cells
                    if cell.row == current_row
                ]
            elif self.direction == consts.FuncArgDirection.VERTICAL:
                cells = [
                    (cell.row, cell.col)
                    for cell in final_cells
                    if cell.col == current_col
                ]
        else:
            cells = [(cell.row, cell.col) for cell in final_cells]
        self.final_cells.extend(cells)


class FuncCell:
    def __init__(
        self,
        row,
        col,
        style,
        initial_value,
        row_height,
        col_width,
        args,
        default_value="",
    ):
        self.row = row
        self.col = col
        self.style = style
        self.initial_value = initial_value
        self.row_height = row_height
        self.col_width = col_width
        self.default_value = default_value
        self.args = {(arg.start_index, arg.end_index): arg for arg in args}
        self.final_args = []

    def move(self, row, col):
        self.row += row
        self.col += col
        for arg in self.final_args:
            arg.final_cells = [
                (arg_row + row, arg_col + col) for arg_row, arg_col in arg.final_cells
            ]

    def finalize_arg(self, arg_key, initial_cell, final_cells):
        arg = self.args[arg_key]
        arg.finalize_cells(self.row, self.col, initial_cell, final_cells)
        if not arg.cells:
            arg = self.args.pop(arg_key)
            self.final_args.append(arg)

    def get_final_value(self):
        str_args = []
        if not self.final_args:
            return self.default_value
        for arg in self.final_args:
            cells = list(sorted(arg.final_cells))
            if not cells:
                return self.default_value
            rectangle_cell_count = (cells[-1][0] - cells[0][0] + 1) * (
                cells[-1][1] - cells[0][1] + 1
            )
            if len(cells) > 1 and len(cells) == rectangle_cell_count:
                start_cell = cells[0]
                start_cell = utils.cell_int_to_str(start_cell[0], start_cell[1])
                end_cell = cells[-1]
                end_cell = utils.cell_int_to_str(end_cell[0], end_cell[1])
                str_value = "{}:{}".format(start_cell, end_cell)
            else:
                str_value = ",".join(
                    utils.cell_int_to_str(cell[0], cell[1]) for cell in cells
                )
            str_args.append((arg.start_index, arg.end_index, str_value))
        value = self.initial_value
        value_parts = []
        value_part_start_index = 0
        for arg in str_args:
            value_parts.append(value[value_part_start_index : arg[0]])
            value_parts.append(arg[2])
            value_part_start_index = arg[1]
        value_parts.append(value[value_part_start_index:])
        return "".join(value_parts)


class CellGroupFinalResult:
    def __init__(self, cells, func_cells, merges, size):
        self.cells = cells
        self.func_cells = func_cells
        self.merges = merges
        self.size = size

    def get_simple_display(self):
        result = [[None] * self.size.width for _ in range(self.size.height)]
        for cell in self.cells:
            result[cell.row][cell.col] = cell.value
        return result


class BaseCellGroup:
    def add_cell(self, row, col, cell):  # pragma: no cover
        raise NotImplementedError()

    def add_cell_group(self, row, col, cell_group):  # pragma: no cover
        raise NotImplementedError()

    @cached_property
    def final_result(self):
        return self.get_final_result()

    def get_final_result(self):  # pragma: no cover
        raise NotImplementedError()

    def get_final_size(self):
        return self.final_result.size

    def get_final_cells(self):
        return self.final_result.cells

    def get_final_func_cells(self):
        return self.final_result.func_cells

    def get_final_merges(self):
        return self.final_result.merges

    def calc_last_cell(self, final_cells, final_func_cells):
        last_cell = (-1, -1)
        for cells in itertools.chain(final_cells.values(), final_func_cells.values()):
            for cell in cells:
                last_cell = (max(last_cell[0], cell.row), max(last_cell[1], cell.col))
        return last_cell


class CellGroup(BaseCellGroup):
    def __init__(self, initial_size: Size):
        self.initial_size = initial_size
        self.cells = []
        self.func_cells = []
        self.cell_groups = {}
        self.merges = []

    def add_merge(self, row, col, rows, cols):
        self.merges.append(Merge(row, col, rows, cols))

    def add_cell(self, cell):
        self.cells.append(cell)

    def add_func_cell(self, cell):
        self.func_cells.append(cell)

    def add_cell_group(self, row, col, cell_group):
        self.cell_groups[(row, col)] = cell_group

    def get_final_result(self):
        row_offsets = [
            [None] * self.initial_size.width
            for _ in range(self.initial_size.height + 1)
        ]
        col_offsets = [
            [None] * self.initial_size.height
            for _ in range(self.initial_size.width + 1)
        ]
        for (row, col), cell_group in self.cell_groups.items():
            final_size = cell_group.get_final_size()
            if final_size.width > cell_group.initial_size.width:
                for i in range(cell_group.initial_size.width):
                    col_offsets[col + i][row] = 0
                col_offsets[col + 1][row] = (
                    final_size.width - cell_group.initial_size.width
                )
            else:
                for i in range(col, col + final_size.width):
                    col_offsets[i][row] = max(col_offsets[i][row] or 0, 0)
                for i in range(
                    col + final_size.width, col + cell_group.initial_size.width
                ):
                    col_offsets[i][row] = -1

            if final_size.height > cell_group.initial_size.height:
                for i in range(cell_group.initial_size.height):
                    row_offsets[row + i][col] = 0
                row_offsets[row + 1][col] = (
                    final_size.height - cell_group.initial_size.height
                )
            else:
                for i in range(row, row + final_size.height):
                    row_offsets[i][col] = max(row_offsets[i][col] or 0, 0)
                for i in range(
                    row + final_size.height, row + cell_group.initial_size.height
                ):
                    row_offsets[i][col] = -1

        for cell in itertools.chain(self.cells, self.func_cells):
            row_offsets[cell.row][cell.col] = max(
                row_offsets[cell.row][cell.col] or 0, 0
            )
            col_offsets[cell.col][cell.row] = max(
                col_offsets[cell.col][cell.row] or 0, 0
            )

        row_offsets = [
            [offset for offset in row if offset is not None] for row in row_offsets
        ]
        row_offsets = [max(row) if row else 0 for row in row_offsets]
        row_offsets = list(itertools.accumulate(row_offsets))

        col_offsets = [
            [offset for offset in col if offset is not None] for col in col_offsets
        ]
        col_offsets = [max(col) if col else 0 for col in col_offsets]
        col_offsets = list(itertools.accumulate(col_offsets))

        final_cells = defaultdict(list)
        final_func_cells = defaultdict(list)
        final_merges = []

        for (cg_row, cg_col), cell_group in self.cell_groups.items():
            for (c_row, c_col), cells in cell_group.get_final_cells().items():
                for cell in cells:
                    cell.move(
                        cg_row + row_offsets[cg_row], cg_col + col_offsets[cg_col]
                    )
                    final_cells[(cg_row + c_row, cg_col + c_col)].append(cell)
            for (c_row, c_col), cells in cell_group.get_final_func_cells().items():
                for cell in cells:
                    cell.move(
                        cg_row + row_offsets[cg_row], cg_col + col_offsets[cg_col]
                    )
                    final_func_cells[(cg_row + c_row, cg_col + c_col)].append(cell)
            for merge in cell_group.get_final_merges():
                merge.move(cg_row + row_offsets[cg_row], cg_col + col_offsets[cg_col])
                final_merges.append(merge)

        for cell in self.cells:
            row, col = cell.row, cell.col
            cell.move(row_offsets[row], col_offsets[col])
            final_cells[(row, col)].append(cell)

        for cell in self.func_cells:
            row, col = cell.row, cell.col
            cell.move(row_offsets[row], col_offsets[col])
            final_func_cells[(row, col)].append(cell)

        for merge in self.merges:
            row, col = merge.row, merge.col
            merge.move(row_offsets[row], col_offsets[col])
            final_merges.append(merge)

        for (row, col), cells in final_func_cells.items():
            finalized_args = []
            for arg_key, arg in cells[0].args.items():
                for cell in arg.cells:
                    final_row = row + cell[0]
                    final_col = col + cell[1]
                    if (
                        final_row >= 0
                        and final_row < self.initial_size.height
                        and final_col >= 0
                        and final_col < self.initial_size.width
                    ):
                        finalized_args.append((arg_key, cell))
            for arg_key, initial_cell in finalized_args:
                for cell in cells:
                    final_row = initial_cell[0] + row
                    final_col = initial_cell[1] + col
                    cell.finalize_arg(
                        arg_key,
                        initial_cell,
                        itertools.chain(
                            final_cells.get((final_row, final_col), []),
                            final_func_cells.get((final_row, final_col), []),
                        ),
                    )

        last_cell = self.calc_last_cell(final_cells, final_func_cells)

        return CellGroupFinalResult(
            cells=final_cells,
            size=Size(width=last_cell[1] + 1, height=last_cell[0] + 1),
            func_cells=final_func_cells,
            merges=final_merges,
        )


class SheetCellGroup(CellGroup):
    def get_final_result(self):
        result = super().get_final_result()
        cells = [cell for _cells in result.cells.values() for cell in _cells]
        func_cells = [cell for _cells in result.func_cells.values() for cell in _cells]
        merges = result.merges

        for cell in itertools.chain(cells, func_cells, merges):
            cell.move(1, 1)

        func_cells = [
            Cell(
                cell.row,
                cell.col,
                cell.style,
                cell.get_final_value(),
                cell.row_height,
                cell.col_width,
            )
            for cell in func_cells
        ]
        cells = cells + func_cells
        cells.sort(key=lambda c: (c.row, c.col))

        return CellGroupFinalResult(
            cells=cells + func_cells,
            func_cells=[],
            merges=merges,
            size=Size(result.size.height + 1, result.size.width + 1),
        )


class LoopCellGroup(BaseCellGroup):
    def __init__(self, initial_size, direction):
        self.initial_size = initial_size
        self.direction = direction
        self.cell_groups = []

    def add_cell_group(self, cell_group):
        self.cell_groups.append(cell_group)

    def get_final_result(self):
        row_offsets = [0]
        col_offsets = [0]
        if self.direction == consts.LoopDirection.DOWN:
            for cell_group in self.cell_groups:
                final_size = cell_group.get_final_size()
                row_offsets.append(final_size.height)
            col_offsets = [0] * len(row_offsets)
        else:
            for cell_group in self.cell_groups:
                final_size = cell_group.get_final_size()
                col_offsets.append(final_size.width)
            row_offsets = [0] * len(col_offsets)

        row_offsets = list(itertools.accumulate(row_offsets))
        col_offsets = list(itertools.accumulate(col_offsets))
        final_cells = defaultdict(list)
        final_func_cells = defaultdict(list)
        final_merges = []
        for index, cell_group in enumerate(self.cell_groups):
            for (row, col), cells in cell_group.get_final_cells().items():
                for cell in cells:
                    cell.move(row_offsets[index], col_offsets[index])
                    final_cells[(row, col)].append(cell)
            for (row, col), cells in cell_group.get_final_func_cells().items():
                for cell in cells:
                    cell.move(row_offsets[index], col_offsets[index])
                    final_func_cells[(row, col)].append(cell)
            for merge in cell_group.get_final_merges():
                merge.move(row_offsets[index], col_offsets[index])
                final_merges.append(merge)

        last_cell = self.calc_last_cell(final_cells, final_func_cells)

        return CellGroupFinalResult(
            cells=final_cells,
            size=Size(width=last_cell[1] + 1, height=last_cell[0] + 1),
            func_cells=final_func_cells,
            merges=final_merges,
        )



import io
from collections import defaultdict

from xlsx_template import nodes


class Symbols:

    BUILTINS = ("range", "len")

    def __init__(self):
        self.symbols = defaultdict(list)

    def declare_ref(self, name):
        self.symbols[name].append("{}_{}".format(name, len(self.symbols[name])))
        return self.symbols[name][-1]

    def undeclare_ref(self, name):
        self.symbols[name].pop(-1)
        if not self.symbols[name]:
            self.symbols.pop(name)

    def add_ref(self, name, real_name):
        self.symbols[name].append(real_name)

    def find_ref(self, name):
        if name in self.symbols:
            return self.symbols[name][-1]
        elif name in self.BUILTINS:
            return name


class CodeGenerator:
    def __init__(self):
        self.indent_count = 0
        self.stream = None
        self.symbols = None
        self.cell_group_level = 0
        self.is_new_line = True

    def generate(self, root_node):
        self.indent_count = 0
        self.stream = io.StringIO()
        self.symbols = Symbols()
        self.is_new_line = True
        self.cell_group_level = 0
        self.generate_for(root_node)
        return self.stream.getvalue()

    def write(self, s):
        if self.is_new_line:
            self.stream.write("    " * self.indent_count)
            self.is_new_line = False
        self.stream.write(s)

    def newline(self):
        self.stream.write("\n")
        self.is_new_line = True

    def write_line(self, s):
        self.write(s)
        self.newline()

    def indent(self):
        self.indent_count += 1

    def unindent(self):
        self.indent_count -= 1

    def generate_for(self, node):
        method_name = "generate_for_{}".format(node.__class__.__name__.lower())
        return getattr(self, method_name)(node)

    def generate_for_sheet(self, sheet_node):
        self.write("sheet = wb.create_sheet(")
        self.generate_for(sheet_node.name)
        self.write(")")
        self.newline()
        size = "cg.Size({}, {})".format(
            sheet_node.last_cell[0] + 1, sheet_node.last_cell[1] + 1
        )
        self.write_line("cell_group_0 = cg.SheetCellGroup({})".format(size))
        self.newline()
        for child in sheet_node.body:
            self.generate_for(child)
        self.newline()
        self.write_line("for f_cell in cell_group_0.get_final_cells():")
        self.indent()
        self.write_line("cell = sheet.cell(f_cell.row, f_cell.col, f_cell.value)")
        self.write_line("if f_cell.style is not None:")
        self.indent()
        self.write_line("cell.style = f_cell.style")
        self.unindent()
        self.write_line("sheet.row_dimensions[f_cell.row].height = f_cell.row_height")
        self.write_line(
            "sheet.column_dimensions[utils.col_int_to_str(f_cell.col)].width = f_cell.col_width"
        )
        self.unindent()
        self.write_line("for m in cell_group_0.get_final_merges():")
        self.indent()
        self.write_line(
            "sheet.merge_cells(start_row=m.row, start_column=m.col, end_row=m.row+m.rows-1, end_column=m.col+m.cols-1)"
        )
        self.unindent()

    def generate_for_sheetloop(self, sheet_loop):
        loop_ref = self.symbols.declare_ref("loop")
        if sheet_loop.name:
            self.symbols.add_ref("{}_loop".format(sheet_loop.name), loop_ref)
        self.write("{} = LoopContext(".format(loop_ref))
        self.generate_for(sheet_loop.items)
        self.write_line(")")
        target_ref = self.symbols.declare_ref(sheet_loop.target)
        self.write_line("for {} in {}:".format(target_ref, loop_ref))
        self.indent()
        self.generate_for_sheet(sheet_loop.sheet)
        self.unindent()
        self.symbols.undeclare_ref("loop")
        self.symbols.undeclare_ref(sheet_loop.target)

    def generate_for_remove(self, remove):
        size = "cg.Size({}, {})".format(remove.height, remove.width)
        self.write_line(
            "cell_group_{} = cg.CellGroup(initial_size={})".format(
                self.cell_group_level + 1, size
            )
        )
        self.write_line(
            "cell_group_{}.add_cell_group({}, {}, cell_group_{})".format(
                self.cell_group_level,
                remove.base_cell[0],
                remove.base_cell[1],
                self.cell_group_level + 1,
            )
        )

    def generate_for_if(self, if_d):
        self.cell_group_level += 1
        size = "cg.Size({}, {})".format(if_d.height, if_d.width)
        self.write_line(
            "cell_group_{} = cg.CellGroup(initial_size={})".format(
                self.cell_group_level, size
            )
        )
        self.write("if ")
        self.generate_for(if_d.condition)
        self.write_line(":")
        self.indent()
        for node in if_d.body:
            self.generate_for(node)
        self.unindent()
        if if_d.else_block:
            self.write_line("else:")
            self.indent()
            for node in if_d.else_block:
                self.generate_for(node)
            self.unindent()
        self.write_line(
            "cell_group_{}.add_cell_group({}, {}, cell_group_{})".format(
                self.cell_group_level - 1,
                if_d.base_cell[0],
                if_d.base_cell[1],
                self.cell_group_level,
            )
        )
        self.cell_group_level -= 1

    def generate_for_cellgroup(self, cell_group):
        self.cell_group_level += 1
        size = "cg.Size({}, {})".format(cell_group.height, cell_group.width)
        self.write_line(
            "cell_group_{} = cg.CellGroup(initial_size={})".format(
                self.cell_group_level, size
            )
        )
        for node in cell_group.body:
            self.generate_for(node)
        self.write_line(
            "cell_group_{}.add_cell_group({}, {}, cell_group_{})".format(
                self.cell_group_level - 1,
                cell_group.base_cell[0],
                cell_group.base_cell[1],
                self.cell_group_level,
            )
        )
        self.cell_group_level -= 1

    def generate_for_cellloop(self, cell_loop):
        self.cell_group_level += 1
        size = "cg.Size({}, {})".format(cell_loop.height, cell_loop.width)
        self.write_line(
            "cell_group_{} = cg.LoopCellGroup(initial_size={}, direction={})".format(
                self.cell_group_level, size, cell_loop.direction
            )
        )
        loop_ref = self.symbols.declare_ref("loop")
        if cell_loop.name:
            self.symbols.add_ref("{}_loop".format(cell_loop.name), loop_ref)
        self.write("{} = LoopContext(".format(loop_ref))
        self.generate_for(cell_loop.items)
        self.write_line(")")
        target_ref = self.symbols.declare_ref(cell_loop.target)
        self.write_line("for {} in {}:".format(target_ref, loop_ref))
        self.indent()
        self.cell_group_level += 1
        self.write_line(
            "cell_group_{} = cg.CellGroup(initial_size={})".format(
                self.cell_group_level, size
            )
        )
        for node in cell_loop.body:
            self.generate_for(node)
        self.write_line(
            "cell_group_{}.add_cell_group(cell_group_{})".format(
                self.cell_group_level - 1, self.cell_group_level
            )
        )
        self.cell_group_level -= 1
        self.unindent()
        self.write_line(
            "cell_group_{}.add_cell_group({}, {}, cell_group_{})".format(
                self.cell_group_level - 1,
                cell_loop.base_cell[0],
                cell_loop.base_cell[1],
                self.cell_group_level,
            )
        )
        self.cell_group_level -= 1
        self.symbols.undeclare_ref("loop")
        self.symbols.undeclare_ref(cell_loop.target)

    def generate_for_funccelloutput(self, func_cell_output):
        self.write_line("fargs = [")
        self.indent()
        for arg in func_cell_output.args:
            self.generate_for(arg)
            self.write_line(",")
        self.unindent()
        self.write_line("]")
        self.write("row_height = ")
        self.generate_for(func_cell_output.row_height)
        self.newline()
        self.write("col_width = ")
        self.generate_for(func_cell_output.col_width)
        self.newline()
        style = (
            None
            if func_cell_output.style is None
            else "'{}'".format(func_cell_output.style)
        )
        self.write(
            "cell = cg.FuncCell({}, {}, {}, '{}', row_height, col_width, fargs, ".format(
                func_cell_output.base_cell[0],
                func_cell_output.base_cell[1],
                style,
                func_cell_output.value,
            )
        )
        if func_cell_output.default_value:
            self.generate_for(func_cell_output.default_value)
        else:
            self.write("None")
        self.write_line(")")
        self.write_line(
            "cell_group_{}.add_func_cell(cell)".format(self.cell_group_level)
        )
        if func_cell_output.merge:
            self.generate_for_merge(func_cell_output)

    def generate_for_funcarg(self, func_arg):
        cells = ", ".join(
            "({}, {})".format(cell[0], cell[1]) for cell in func_arg.cells
        )
        direction = "None" if func_arg.direction is None else str(func_arg.direction)
        self.write(
            "cg.FuncArg({}, {}, [{}], {})".format(
                func_arg.start_index, func_arg.end_index, cells, direction
            )
        )

    def generate_for_merge(self, cell_output):
        self.write(
            "cell_group_{}.add_merge({}, {}, ".format(
                self.cell_group_level,
                cell_output.base_cell[0],
                cell_output.base_cell[1],
            )
        )
        if cell_output.merge.rows:
            self.generate_for(cell_output.merge.rows)
        else:
            self.write("1")
        self.write(", ")
        if cell_output.merge.cols:
            self.generate_for(cell_output.merge.cols)
        else:
            self.write("1")
        self.write_line(")")

    def generate_for_getattr(self, get_attr):
        self.write("env.get_attr(")
        self.generate_for(get_attr.obj)
        self.write(', "{}")'.format(get_attr.attr_name))

    def generate_for_getitem(self, get_item):
        self.write("env.get_item(")
        self.generate_for(get_item.obj)
        self.write(", ")
        self.generate_for(get_item.key)
        self.write(")")

    def generate_for_celloutput(self, cell_output):
        style = None if cell_output.style is None else "'{}'".format(cell_output.style)
        self.write("row_height = ")
        self.generate_for(cell_output.row_height)
        self.newline()
        self.write("col_width = ")
        self.generate_for(cell_output.col_width)
        self.newline()
        self.write(
            "cell = cg.Cell({}, {}, {}, ".format(
                cell_output.base_cell[0], cell_output.base_cell[1], style
            )
        )
        if cell_output.value:
            self.generate_for(cell_output.value)
        else:
            self.write("None")
        self.write(", row_height, col_width)")
        self.newline()
        self.write_line("cell_group_{}.add_cell(cell)".format(self.cell_group_level))
        if cell_output.merge:
            self.generate_for_merge(cell_output)

    def generate_for_call(self, call_node):
        self.generate_for(call_node.obj)
        self.write("(")
        for arg in call_node.args:
            self.generate_for(arg)
            self.write(", ")
        self.write(")")

    def generate_for_arg(self, arg):
        self.generate_for(arg.value)

    def generate_for_value(self, value):
        for child in value.body[:-1]:
            self.generate_for(child)
            self.write(" + ")
        self.generate_for(value.body[-1])

    def generate_for_const(self, const):
        self.write(repr(const.value))

    def generate_for_strconst(self, const):
        self.write(const.value)

    def generate_for_filter(self, filter_):
        self.write('env.filters["{}"]'.format(filter_.name))
        self.write("(")
        self.generate_for(filter_.obj)
        self.write(", ")
        for arg in filter_.args:
            self.generate_for(arg)
            self.write(", ")
        self.write(")")

    def generate_for_kwarg(self, kwarg):
        self.write("{}=".format(kwarg.name))
        self.generate_for(kwarg.value)

    def generate_for_tostr(self, to_str):
        self.write("str(")
        self.generate_for(to_str.value)
        self.write(")")

    def generate_for_var(self, var):
        ref = self.symbols.find_ref(var.name)
        if ref is not None:
            res = ref
        else:
            res = 'ctx.resolve("{}")'.format(var.name)
        self.write(res)

    def generate_for_template(self, template_node):
        self.write_line("import io")
        self.write_line("import datetime")
        self.newline()
        self.write_line("import openpyxl")
        self.newline()
        self.write_line(
            "from xlsx_template.runtime import cell_groups as cg, LoopContext"
        )
        self.write_line(
            "from xlsx_template.consts import LoopDirection, FuncArgDirection"
        )
        self.write_line("from xlsx_template import utils")
        self.newline()
        self.newline()
        self.write_line("def root(context, styles, env):")
        self.indent()
        self.write_line("ctx = context")
        self.write_line("wb = openpyxl.Workbook()")
        self.write_line("del wb['Sheet']")
        self.write_line("for style in styles:")
        self.indent()
        self.write_line("wb.add_named_style(style)")
        self.unindent()
        self.newline()
        for child_node in template_node.body:
            self.generate_for(child_node)
            self.newline()
        self.write_line("buf = io.BytesIO()")
        self.write_line("wb.save(buf)")
        self.write_line("return buf.getvalue()")
        self.unindent()
        self.write_line("")



class ParseError(Exception):
    def __init__(self, msg, source_hint, orig_exc=None):
        self.msg = msg
        self.source_hint = source_hint
        self.orig_exc = orig_exc


class TemplateRuntimeException(RuntimeError):
    def __init__(self, msg, orig_exception=None):
        super().__init__(msg)
        self.orig_exception = orig_exception



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xlsx_template', 'xlsx_template.runtime']

package_data = \
{'': ['*']}

install_requires = \
['cached-property>=1.5.2,<2.0.0',
 'openpyxl>=3.0.5,<4.0.0',
 'pyparsing>=2.4.7,<3.0.0']

setup_kwargs = {
    'name': 'xlsx-template',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Sergei Sinitsyn',
    'author_email': 'sinitsinsv@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import string
import collections


def col_str_to_int(col):
    final_column = 0
    for letter in col:
        final_column = final_column * 26 + string.ascii_uppercase.index(letter) + 1
    return final_column


def cell_str_to_int(row, col):
    row = int(row)
    return (row, col_str_to_int(col))


def cell_int_to_str(row, col):
    return col_int_to_str(col) + str(row)


def col_int_to_str(col):
    res = ""
    while col > 0:
        col, reminder = divmod(col, 26)
        if reminder == 0:
            reminder = 26
            col -= 1
        res = string.ascii_uppercase[reminder - 1] + res
    return res


# Cell = collections.namedtuple("Cell", "row,col")

# class Cell:
#     def __init__(self, row, col):
#         self.row = row
#         self.col = col

#     def adjust(self, row, col):
#         self.row += row
#         self.col += col

#     def __eq__(self, other):
#         return self.row == other.row and self.col == other.col

#     def __hash__(self):
#         return (self.row, self.col).__hash__()

# def __str__(self):
#     return cell_int_to_str(self.row, self.col)


# class CellRange:
#     def __init__(self, start_cell, end_cell):
#         self.start_cell = start_cell
#         self.end_cell = end_cell



def default_if_none(value, default_value):
    return default_value if value is None else value


def yes_no(value, yes_value, no_value):
    if value:
        return yes_value
    else:
        return no_value
